// Define IS_STANDALONE as 1 for standalone application, i.e. that'll work w/o DOS,
// Define IS_STANDALONE as 0 for standalone application, i.e. that'll work with DOS:
#define IS_STANDALONE   01

// The application must call stdio_DoneFATFS() before it exists in order to
// flush all unwritten/buffered data to the disk.
// If needed, use stdio_FlushFATFS() to flush all unwritten/buffered data
// to the disk in the middle of the application operation.

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <conio.h>
#include <dos.h>
#include <time.h>
#include <errno.h>
#include <sys/stat.h>

#if defined(__TURBOC__)
#include <dir.h>
#elif defined(__WATCOMC__)
#include <direct.h>
#endif

#include "stdio/mydirent.h"
#include "MtchIter.h"
#include "fcpy.h"

#if IS_STANDALONE
  #ifdef __cplusplus
  extern "C" {
  #endif
extern int stdio_InitFATFS();
extern int stdio_DoneFATFS();
extern int stdio_FlushFATFS();
  #ifdef __cplusplus
  }
  #endif
#else
  #define stdio_InitFATFS()
  #define stdio_DoneFATFS()
  #define stdio_FlushFATFS()
#endif

#define CMDLINE_MAX_LEN 79

char aCmdLine[CMDLINE_MAX_LEN+1];

typedef enum tCmd {
  c_DIR,
  c_COPY,
  c_DEL,
  c_ATTRIB,
  c_DISK,
  c_CD,
  c_MD,
  c_RD,
  c_PWD,
  c_TIME,
  c_MEM,
  c_HELP,
  c_EXIT,
  c_LAST_CMD=c_EXIT
} tCmd;

const char* aCommands[]=
{
  "DIR",
  "COPY",
  "DEL",
  "ATTRIB",
  "D:",
  "CD",
  "MD",
  "RD",
  "PWD",
  "TIME",
  "MEM",
  "HELP",
  "EXIT"
};

const char* aCommandsHelp[]=
{
  "DIR [path] shows contents of a directory",
  "COPY <file(s)> [path] copies one or more files",
  "DEL <file(s)> deletes one or more files (except read-only files)",
  "ATTRIB <+R | -R> <file(s)> sets/resets the read-only attribute of file(s)",
  "D: changes the current disk",
  "CD <path> changes the current directory",
  "MD <path> creates a directory",
  "RD <path> removes a directory",
  "PWD shows the full name of the current directory",
  "TIME shows the current date and time",
  "MEM shows the amount of free heap memory",
  "HELP [command] shows help for a command",
  "EXIT exits the program"
};

size_t GetFreeHeapSize()
{
  size_t sz;
  void* p;
  int m, i;

  sz = 1;
  m = 0;
  for (;;)
  {
    p = malloc (sz);
    if (p == NULL)
      break;
    free (p);
    m++;
    sz <<= 1;
  }
  if (!m)
    return 0;
  m--;
  sz = ((size_t)1) << m;

  for (i=m-1; i>=0; i--)
  {
    size_t s = ((size_t)1) << i;
    p = malloc (sz | s);
    if (p != NULL)
    {
      free (p);
      sz |= s;
    }
  }

  return sz;
}

static int IsItDir (const char* path)
{
  DIR *d = opendir ((char*)path);
  if (d == NULL)
    return 0;
  closedir (d);
  return 1;
}

void cd (const char* pPath)
{
  if (pPath == NULL)
  {
    printf ("Path not specified\n");
    return;
  }

  errno = 0;
  if (chdir (pPath))
    printf ("Error\n");
}

void cdisk (int disk)
{
  errno = 0;
#if defined(__TURBOC__) && (__TURBOC__ <= 0x0296)
  setdisk (disk);
  if (getdisk() != disk)
    printf ("Error\n");
#else
  {
  unsigned ndisks;
  _dos_setdrive (1+disk, &ndisks);
  _dos_getdrive (&ndisks);
  if (1+disk != ndisks)
    printf ("Error\n");
  }
#endif
}

void md (const char* pPath)
{
  if (pPath == NULL)
  {
    printf ("Path not specified\n");
    return;
  }

  errno = 0;
  if (mkdir (pPath))
    printf ("Error\n");
}

void rd (const char* pPath)
{
  if (pPath == NULL)
  {
    printf ("Path not specified\n");
    return;
  }

  errno = 0;
  if (rmdir (pPath))
    printf ("Error\n");
}

void pwd()
{
  char s[192];

  errno = 0;
  if (getcwd (s, sizeof(s)) == NULL)
  {
    printf ("\nPathname is too long\n");
    return;
  }
  printf ("%s\n", s);
}

unsigned FilesFound;
unsigned long FilesFoundSize;

void PrintFileOperationStatistics()
{
  printf ("%9u file(s)        %lu bytes\n", FilesFound, FilesFoundSize);
}

static void IteratedDirFxn (const char* pName, const char* pNameInDir,
                            struct stat* pStatBuf)
{
  struct tm* pBrokenDownTime;

  (void)pName;

  FilesFound++;

  printf
  (
    "%- 13s ",
    pNameInDir
  );

  pBrokenDownTime = localtime (&pStatBuf->st_mtime);

  if (pStatBuf->st_mode & S_IFDIR)
    printf ("<DIR>     ");
  else
  {
    printf ("%10lu", pStatBuf->st_size);
    FilesFoundSize += pStatBuf->st_size;
  }

  printf
  (
    " %04u-%02u-%02u %02u:%02u:%02u %c%c%c",
    pBrokenDownTime->tm_year+1900,
    pBrokenDownTime->tm_mon+1,
    pBrokenDownTime->tm_mday,
    pBrokenDownTime->tm_hour,
    pBrokenDownTime->tm_min,
    pBrokenDownTime->tm_sec,
    "-r"[(pStatBuf->st_mode&S_IREAD)!=0],
    "-w"[(pStatBuf->st_mode&S_IWRITE)!=0],
    "-x"[(pStatBuf->st_mode&S_IEXEC)!=0]
  );

  printf ("\n");
}

void dir (const char* pPath)
{
  int res;

  if ((pPath == NULL) || (!*pPath))
  {
    dir (".");
    return;
  }

  printf ("Name                Size YYYY-MM-DD HH:MM:SS RWX\n");

  FilesFound = 0;
  FilesFoundSize = 0;

  errno = 0;
  res = MatchIterator (pPath, 192, &IteratedDirFxn);

  if (res < 0)
  {
    printf ("Error\n");
  }
  else if (res > 0)
  {
    char a[1+CMDLINE_MAX_LEN+4];
    strcpy (a, pPath);
    switch (pPath[strlen(pPath)-1])
    {
      case ':':
      case '\\':
        break;
      default:
        strcat (a, "\\");
        break;
    }

    strcat (a, "*.*");
    errno = 0;
    res = MatchIterator (a, 192, &IteratedDirFxn);
    if (res < 0)
    {
      printf ("Error\n");
    }
  }

  PrintFileOperationStatistics();
}

static void IteratedDelFxn (const char* pName, const char* pNameInDir,
                            struct stat* pStatBuf)
{
  (void)pName;
  (void)pStatBuf;

  if (!(pStatBuf->st_mode & S_IFREG))
    return;

  printf
  (
    "%s",
    pNameInDir
  );

  errno = 0;
  if (remove (pName))
  {
    printf (" Error\n");
  }
  else
  {
    printf ("\n");
    FilesFound++;
    FilesFoundSize += pStatBuf->st_size;
  }
}

void del (const char* pPath)
{
  int res;

  if (pPath == NULL)
  {
    printf ("File(s) not specified\n");
    return;
  }

  FilesFound = 0;
  FilesFoundSize = 0;

  errno = 0;
  res = MatchIterator (pPath, 192, &IteratedDelFxn);

  if (res)
  {
    printf ("Error\n");
  }

  PrintFileOperationStatistics();
}

char aCopyDstPath[192];
int IsDirCopyDstPath;

static void IteratedCopyFxn (const char* pName, const char* pNameInDir,
                             struct stat* pStatBuf)
{
  char aDstName[192];

  if (!(pStatBuf->st_mode & S_IFREG))
    return;

  printf
  (
    "%s",
    pNameInDir
  );

  errno = 0;
  strcpy (aDstName, aCopyDstPath);
  if (IsDirCopyDstPath)
    strcat (aDstName, pNameInDir);
  if (fcpy (pName, aDstName, NULL, 4096))
  {
    printf (" Error\n");
  }
  else
  {
    printf ("\n");
    FilesFound++;
    FilesFoundSize += pStatBuf->st_size;
  }
}

void copy (const char* pPath, const char* pDstPath)
{
  int res;

  if (pPath == NULL)
  {
    printf ("File(s) not specified\n");
    return;
  }

  if (pDstPath == NULL)
  {
    copy (pPath, ".");
    return;
  }

  IsDirCopyDstPath = IsItDir (pDstPath);
  strcpy (aCopyDstPath, pDstPath);
  if (IsDirCopyDstPath)
  {
    switch (aCopyDstPath[strlen(aCopyDstPath)-1])
    {
      case ':':
      case '\\':
        break;
      default:
        strcat (aCopyDstPath, "\\");
        break;
    }
  }

  FilesFound = 0;
  FilesFoundSize = 0;

  errno = 0;
  res = MatchIterator (pPath, 192, &IteratedCopyFxn);

  if (res)
  {
    printf ("Error\n");
  }

  PrintFileOperationStatistics();
}

// Old Turbo/Borland C/C++ doesn't have the S_I?USR defines:
#if !defined(S_IRUSR)
  #define S_IRUSR S_IREAD
  #define S_IWUSR S_IWRITE
  #define S_IXUSR S_IEXEC
#endif

int AttribSetNeeded;

static void IteratedAttribFxn (const char* pName, const char* pNameInDir,
                               struct stat* pStatBuf)
{
  if (!(pStatBuf->st_mode & S_IFREG))
    return;

  printf
  (
    "%s",
    pNameInDir
  );

  errno = 0;
  // set/reset writable (!read-only) attribute:
  pStatBuf->st_mode = (pStatBuf->st_mode & ~S_IWUSR) |
                      (AttribSetNeeded?0:S_IWUSR);
  if (chmod (pName, pStatBuf->st_mode))
  {
    printf (" Error\n");
  }
  else
  {
    printf ("\n");
    FilesFound++;
    FilesFoundSize += pStatBuf->st_size;
  }
}

void attrib (const char* pAction, const char* pPath)
{
  int res;
  enum {a_NONE, a_SET, a_RESET} action = a_NONE;

  if (pAction == NULL)
  {
    printf ("Action not specified, must be either +R or -R\n");
    return;
  }

  if ((!strcmp(pAction,"+R")) || (!strcmp(pAction,"+r")))
    action = a_SET;
  if ((!strcmp(pAction,"-R")) || (!strcmp(pAction,"-r")))
    action = a_RESET;
  if (action == a_NONE)
  {
    printf ("Wrong action specified, must be either +R or -R\n");
    return;
  }

  if (pPath == NULL)
  {
    printf ("File(s) not specified\n");
    return;
  }

  FilesFound = 0;
  FilesFoundSize = 0;
  AttribSetNeeded = action == a_SET;

  errno = 0;
  res = MatchIterator (pPath, 192, &IteratedAttribFxn);

  if (res)
  {
    printf ("Error\n");
  }

  PrintFileOperationStatistics();
}

void _time()
{
  time_t t;
  struct tm *pt;

  time (&t);
  pt = localtime (&t);
  printf
  (
    "YYYY-MM-DD HH:MM:SS\n"
    "%4u-%02u-%02u %02u:%02u:%02u\n",
    pt->tm_year+1900, pt->tm_mon+1, pt->tm_mday,
    pt->tm_hour, pt->tm_min, pt->tm_sec
  );
}

void mem()
{
  printf ("Free heap memory (64KB max): %lu bytes\n", (unsigned long)GetFreeHeapSize());
}

int ParseAndProcessCmd (char* pCmdLine)
{
  int i, cmd = -1;
#define MAX_CMD_TOKENS  3
  char* apToken[MAX_CMD_TOKENS]={NULL,NULL,NULL};
  int ArgCnt = 0;

  strupr (pCmdLine);

  apToken[0] = strtok (pCmdLine, " \t");

  for (i=0; i<=c_LAST_CMD; i++)
    if ((strlen(aCommands[i])) && (!strncmp (aCommands[i], pCmdLine, strlen(aCommands[i]))))
    {
      cmd = i;
      break;
    }

  if
  (
    (cmd < 0) &&
    (apToken[0] != NULL) &&
    (apToken[0][1]==':') &&
    (!apToken[0][2])
  )
    cmd = c_DISK;

  if (cmd < 0)
  {
    printf ("Unknown command, try HELP.\n");
    return 0;
  }

  for (i=1; i<MAX_CMD_TOKENS; i++)
  {
    apToken[i] = strtok (NULL, " \t");
    if (apToken[i] != NULL)
      ArgCnt++;
    else
      break;
  }

  switch (cmd)
  {
    case c_DISK:
      cdisk (apToken[0][0]-'A');
      return 0;

    case c_CD:
      cd (apToken[1]);
      return 0;

    case c_PWD:
      pwd();
      return 0;

    case c_DIR:
      dir (apToken[1]);
      return 0;

    case c_MD:
      md (apToken[1]);
      return 0;

    case c_RD:
      rd (apToken[1]);
      return 0;

    case c_DEL:
      del (apToken[1]);
      return 0;

    case c_COPY:
      copy (apToken[1], apToken[2]);
      return 0;

    case c_ATTRIB:
      attrib (apToken[1], apToken[2]);
      return 0;

    case c_TIME:
      _time();
      return 0;

    case c_MEM:
      mem();
      return 0;

    case c_HELP:
      if (apToken[1] != NULL)
      {
        for (cmd=-1,i=0; i<=c_LAST_CMD; i++)
          if (!strncmp (aCommands[i], apToken[1], strlen(aCommands[i])))
          {
            cmd = i;
            break;
          }

        if (cmd < 0)
        {
          printf ("Unknown command, try HELP.\n");
          return 0;
        }

        printf ("%s\n", aCommandsHelp[cmd]);
      }
      else
      {
        printf ("Available commands:\n");
        for (i=0; i<=c_LAST_CMD; i++)
          printf ("%s\n", aCommands[i]);
      }
      return 0;

    case c_EXIT:
      return 1;

    default:
      return 0;
  }
}

int main()
{
  aCmdLine[0] = 0;

  stdio_InitFATFS();

  printf ("FAT Commander. Type HELP or EXIT...\n");

  for(;;)
  {
    char ch[2];
    ch[0] = getch();
    ch[1] = 0;

    switch (ch[0])
    {
      case '\r':
      break;

      case '\b':
        if (aCmdLine[0])
        {
          aCmdLine[strlen(aCmdLine)-1] = 0;
          printf ("\b \b");
        }
      break;

      default:
        if (strlen (aCmdLine) <= CMDLINE_MAX_LEN)
          strcat (aCmdLine, ch);
        printf ("%c", ch[0]);
      break;
    }

    if (ch[0] != '\r')
      continue;

    printf ("\n");
    if (ParseAndProcessCmd (aCmdLine))
      break;
    stdio_FlushFATFS(); // flush everything after a command
    aCmdLine[0] = 0;
  }

  stdio_DoneFATFS();

  return 0;
}
