#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <errno.h>
#include <time.h>
#include <sys/stat.h>
#include "FAT.h"

#ifdef __cplusplus
extern "C" {
#endif
extern tFATFS stdio_FATFS;
extern int stdio_IsFATFSInit;
extern int stdio_Err2ErrNo (tIORes IORes);
#ifdef __cplusplus
}
#endif

typedef struct tFATFILE {
  tFATFile FATFile;
  int IsSearchable;
  struct tFATFILE* pPrev;
} tFATFILE;

tFATFILE* stdio_pLastFile = NULL;

static tFATFILE* AllocateFileStructure()
{
  tFATFILE* pFATFILE;

  pFATFILE = (tFATFILE*) malloc (sizeof(tFATFILE));

  if (pFATFILE != NULL)
  {
    pFATFILE->IsSearchable = 0;
    pFATFILE->pPrev = stdio_pLastFile;
    stdio_pLastFile = pFATFILE;
  }

  return pFATFILE;
}

static void FreeFileStructure (tFATFILE* pFATFILE)
{
  if (pFATFILE == NULL)
    return;

  if (pFATFILE == stdio_pLastFile)
  {
    stdio_pLastFile = pFATFILE->pPrev;
    free (pFATFILE);
  }
  else
  {
    tFATFILE* p;
    for (p=stdio_pLastFile; p!=NULL; p=p->pPrev)
      if (p->pPrev == pFATFILE)
      {
        p->pPrev = pFATFILE->pPrev;
        free (pFATFILE);
        break;
      }
  }
}

long _fsize (FILE *stream);

int chdir (const char *dirname)
{
  tIORes IORes;
  if (!stdio_IsFATFSInit)
    return -1;

  if ((dirname == NULL) || (!*dirname))
  {
    errno = EINVAL;
    return -1;
  }

  FAT_chdir (&stdio_FATFS, &IORes, dirname, 0);

  if (IORes != ior_OK)
  {
    errno = stdio_Err2ErrNo (IORes);
    return -1;
  }

  return 0;
}

char* getcwd (char *buffer,
#if defined(__TURBOC__)
int
#else
size_t
#endif
max)
{
  int IsAllocated = 0;
  tIORes IORes;

  if (!stdio_IsFATFSInit)
    return NULL;

  if (max <= 0)
  {
    errno = EINVAL;
    return NULL;
  }

  if (buffer == NULL)
  {
    buffer = (char*) malloc (max);
    if (buffer == NULL)
    {
      errno = ENOMEM;
      return NULL;
    }
    IsAllocated = 1;
  }

  if (FAT_getcwd (&stdio_FATFS, &IORes, buffer, max) == NULL)
  {
    if (IsAllocated)
      free (buffer);
    buffer = NULL;
  }

  if (IORes != ior_OK)
    errno = stdio_Err2ErrNo (IORes);

  return buffer;
}

FILE* fopen (const char *filename, const char *mode)
{
  tFATFILE* pFATFILE;
  tIORes IORes;

  if (!stdio_IsFATFSInit)
    return NULL;

  if ((filename == NULL) || (!*filename) || (mode == NULL) || (!*mode))
  {
    errno = EINVAL;
    return NULL;
  }

  pFATFILE = AllocateFileStructure();

  if (pFATFILE == NULL)
  {
    errno = ENOMEM;
    return NULL;
  }

  if (!FAT_fopen (&stdio_FATFS, &IORes, &pFATFILE->FATFile, filename, mode))
  {
    FreeFileStructure (pFATFILE);
    pFATFILE = NULL;
  }
  else
  {
    pFATFILE->IsSearchable = 1;
  }

  if (IORes != ior_OK)
    errno = stdio_Err2ErrNo (IORes);

  return (FILE*)pFATFILE;
}

int fclose (FILE *stream)
{
  tFATFILE* pFATFILE = (tFATFILE*)stream;
  tIORes IORes;

  if (!stdio_IsFATFSInit)
    return EOF;

  if (pFATFILE == NULL)
  {
    errno = EINVAL;
    return EOF;
  }

  FAT_fclose (&stdio_FATFS, &IORes, &pFATFILE->FATFile);
  FreeFileStructure (pFATFILE);

  if (IORes != ior_OK)
  {
    errno = stdio_Err2ErrNo (IORes);
    return EOF;
  }

  return 0;
}

int fcloseall()
{
  int cnt = 0, res = ~EOF;

  if (!stdio_IsFATFSInit)
    return EOF;

  while (stdio_pLastFile != NULL)
    if (!fclose ((FILE*)stdio_pLastFile))
      cnt++;
    else
      res = EOF;

  if (res == ~EOF)
    res = cnt;

  return res;
}

int fseek (FILE *stream, long offset, int whence)
{
  tFATFILE* pFATFILE = (tFATFILE*)stream;
  long curpos, newpos = 0, fsize;
  tIORes IORes;

  if (!stdio_IsFATFSInit)
    return -1;

  curpos = ftell (stream);
  if (curpos < 0)
    return -1;

  fsize = _fsize (stream);
  if (fsize < 0)
    return -1;

  switch (whence)
  {
    case SEEK_SET:
      newpos = offset;
    break;
    case SEEK_CUR:
      newpos = curpos + offset;
    break;
    case SEEK_END:
      newpos = fsize + offset;
    break;
    default:
      errno = EINVAL;
      return -1;
  };

  if (newpos < 0)
  {
    errno = ERANGE;
    return -1;
  }

  FAT_fseek (&stdio_FATFS, &IORes, &pFATFILE->FATFile, newpos);

  if (IORes != ior_OK)
  {
    errno = stdio_Err2ErrNo (IORes);
    return -1;
  }

  return 0;
}

void rewind (FILE *stream)
{
  tFATFILE* pFATFILE = (tFATFILE*)stream;
  tIORes IORes;

  if (!stdio_IsFATFSInit)
    return;

  if (pFATFILE == NULL)
  {
    errno = EINVAL;
    return;
  }

  FAT_rewind (&stdio_FATFS, &IORes, &pFATFILE->FATFile);

  if (IORes != ior_OK)
    errno = stdio_Err2ErrNo (IORes);
}

long ftell (FILE *stream)
{
  tFATFILE* pFATFILE = (tFATFILE*)stream;
  uint32 Position;
  tIORes IORes;

  if (!stdio_IsFATFSInit)
    return -1;

  if (pFATFILE == NULL)
  {
    errno = EINVAL;
    return -1;
  }

  FAT_ftell (&stdio_FATFS, &IORes, &pFATFILE->FATFile, &Position);

  if (IORes != ior_OK)
  {
    errno = stdio_Err2ErrNo (IORes);
    return -1;
  }

  return Position;
}

// feof() is normally defined as a macro...:
#ifdef feof
  #undef feof
#endif
int feof (FILE *stream)
{
  tFATFILE* pFATFILE = (tFATFILE*)stream;
  int IsEOF;
  tIORes IORes;

  if (!stdio_IsFATFSInit)
    return 1;

  if (pFATFILE == NULL)
  {
    errno = EINVAL;
    return 1;
  }

  FAT_feof (&stdio_FATFS, &IORes, &pFATFILE->FATFile, &IsEOF);

  if (IORes != ior_OK)
  {
    errno = stdio_Err2ErrNo (IORes);
    return 1;
  }

  return IsEOF;
}

long _fsize (FILE *stream)
{
  tFATFILE* pFATFILE = (tFATFILE*)stream;
  uint32 size;
  tIORes IORes;

  if (!stdio_IsFATFSInit)
    return -1;

  if (pFATFILE == NULL)
  {
    errno = EINVAL;
    return -1;
  }

  FAT_fsize (&stdio_FATFS, &IORes, &pFATFILE->FATFile, &size);

  if (IORes != ior_OK)
  {
    errno = stdio_Err2ErrNo (IORes);
    return -1;
  }

  return size;
}

int fgetpos (FILE *stream, fpos_t *pos)
{
  long p;

  if (!stdio_IsFATFSInit)
    return 1;

  p = ftell (stream);
  if (p < 0)
    return 1;

  if (pos == NULL)
  {
    errno = EINVAL;
    return 1;
  }

  *pos = p;
  return 0;
}

int fsetpos (FILE *stream, const fpos_t *pos)
{
  if (!stdio_IsFATFSInit)
    return 1;

  if (pos == NULL)
  {
    errno = EINVAL;
    return 1;
  }

  return fseek (stream, *pos, SEEK_SET) != 0;
}

size_t fread (void *ptr, size_t size, size_t n, FILE *stream)
{
  tFATFILE* pFATFILE = (tFATFILE*)stream;
  size_t res;
  tIORes IORes;

  if (!stdio_IsFATFSInit)
    return 0;

  if ((pFATFILE == NULL) || (ptr == NULL))
  {
    errno = EINVAL;
    return 0;
  }

  res = FAT_fread (&stdio_FATFS, &IORes, ptr, size, n, &pFATFILE->FATFile);

  if (IORes != ior_OK)
  {
    errno = stdio_Err2ErrNo (IORes);
    return 0;
  }

  return res;
}

size_t fwrite (const void *ptr, size_t size, size_t n, FILE *stream)
{
  tFATFILE* pFATFILE = (tFATFILE*)stream;
  size_t res;
  tIORes IORes;

  if (!stdio_IsFATFSInit)
    return 0;

  if ((pFATFILE == NULL) || (ptr == NULL))
  {
    errno = EINVAL;
    return 0;
  }

  res = FAT_fwrite (&stdio_FATFS, &IORes, ptr, size, n, &pFATFILE->FATFile);

  if (IORes != ior_OK)
  {
    errno = stdio_Err2ErrNo (IORes);
    return 0;
  }

  return res;
}

int fgetc (FILE *stream)
{
  unsigned char c;

  if (!stdio_IsFATFSInit)
    return EOF;

  if (feof(stream))
    return EOF;
  if (fread (&c, 1, 1, stream))
    return c;
  else
    return EOF;
}

int fputc (int ch, FILE *stream)
{
  unsigned char c = (unsigned char)ch;

  if (!stdio_IsFATFSInit)
    return EOF;

  if (fwrite (&c, 1, 1, stream))
    return c;
  else
    return EOF;
}

char* fgets (char*s, int n, FILE *stream)
{
  int c=0;
  char *p=s;

  if (!stdio_IsFATFSInit)
    return NULL;

  if (s == NULL)
  {
    errno = EINVAL;
    return NULL;
  }

  while ((--n > 0) && ((c=fgetc(stream)) != EOF))
  {
    *p++ = c;
    if (c == '\n')
      break;
  }

  if ((c == EOF) && (p == s))
    return NULL;

  *p++ = 0;
  return s;
}

int fputs (const char*s, FILE *stream)
{
  size_t l;

  if (!stdio_IsFATFSInit)
    return EOF;

  if ((s == NULL) || (!*s))
  {
    errno = EINVAL;
    return EOF;
  }

  l = strlen (s);

  if (fwrite (s, 1, l, stream) == l)
    return (unsigned char)s[l-1];
  else
    return EOF;
}

int fflush (FILE *stream)
{
  tFATFILE* pFATFILE = (tFATFILE*)stream;
  tIORes IORes;

  if (!stdio_IsFATFSInit)
    return EOF;

  if (pFATFILE == NULL)
  {
    errno = EINVAL;
    return EOF;
  }

  FAT_fflush (&stdio_FATFS, &IORes, &pFATFILE->FATFile);

  if (IORes != ior_OK)
  {
    errno = stdio_Err2ErrNo (IORes);
    return EOF;
  }

  return 0;
}

int flushall()
{
  int res = 0;
  tFATFILE* pFATFILE;

  if (!stdio_IsFATFSInit)
    return 0;

  for (pFATFILE=stdio_pLastFile; pFATFILE!=NULL; pFATFILE=pFATFILE->pPrev,res++)
    fflush ((FILE*)pFATFILE);

  return res;
}

int unlink (const char* filename)
{
  tIORes IORes;

  if (!stdio_IsFATFSInit)
    return -1;

  if ((filename == NULL) || (!*filename))
  {
    errno = EINVAL;
    return -1;
  }

  FAT_unlink (&stdio_FATFS, &IORes, filename);

  if (IORes != ior_OK)
  {
    errno = stdio_Err2ErrNo (IORes);
    return -1;
  }

  return 0;
}

// remove() is normally defined as a macro...:
#ifdef remove
  #undef remove
#endif
int remove (const char* filename)
{
  return unlink (filename);
}

int mkdir (const char* path
#if !defined(__TURBOC__) && !defined(__WATCOMC__)
, int mode
#endif
)
{
  tIORes IORes;

  if (!stdio_IsFATFSInit)
    return -1;

  if ((path == NULL) || (!*path))
  {
    errno = EINVAL;
    return -1;
  }

#if !defined(__TURBOC__) && !defined(__WATCOMC__)
  FAT_mkdir (&stdio_FATFS, &IORes, path, mode);
#else
  FAT_mkdir (&stdio_FATFS, &IORes, path, 0);
#endif

  if (IORes != ior_OK)
  {
    errno = stdio_Err2ErrNo (IORes);
    return -1;
  }

  return 0;
}

int rmdir (const char* path)
{
  tIORes IORes;

  if (!stdio_IsFATFSInit)
    return -1;

  if ((path == NULL) || (!*path))
  {
    errno = EINVAL;
    return -1;
  }

  FAT_rmdir (&stdio_FATFS, &IORes, path);

  if (IORes != ior_OK)
  {
    errno = stdio_Err2ErrNo (IORes);
    return -1;
  }

  return 0;
}

void _dos_getdrive (unsigned *drive)
{
  tIORes IORes;

  if (!stdio_IsFATFSInit)
    return;

  FAT_dos_getdrive (&stdio_FATFS, &IORes, drive);

  if (IORes != ior_OK)
    errno = stdio_Err2ErrNo (IORes);
}

void _dos_setdrive (unsigned drive, unsigned *ndrives)
{
  tIORes IORes;

  if (!stdio_IsFATFSInit)
    return;

  FAT_dos_setdrive (&stdio_FATFS, &IORes, drive, ndrives);

  if (IORes != ior_OK)
    errno = stdio_Err2ErrNo (IORes);
}

int getdisk()
{
  unsigned disk;
  _dos_getdrive (&disk);
  return disk-1;
}

int setdisk (int disk)
{
  unsigned disks;
  _dos_setdrive (1+disk, &disks);
  return disks;
}

// Old Turbo/Borland C/C++ doesn't have the ?_OK defines:
#if !defined(__TURBOC__)
  #include <unistd.h>
#elif !defined(F_OK)
  #define F_OK 0
  #define X_OK 1
  #define W_OK 2
  #define R_OK 4
#endif

int access (const char *filename, int amode)
{
  tIORes IORes;
  tAccessMode mode = (tAccessMode)0;

  if (!stdio_IsFATFSInit)
    return -1;

  if (amode == F_OK)
    mode = am_F_OK;
  else if (amode & X_OK)
    mode |= am_X_OK;
  else if (amode & W_OK)
    mode |= am_W_OK;
  else if (amode & R_OK)
    mode |= am_R_OK;

  FAT_access (&stdio_FATFS, &IORes, filename, mode);

  if (IORes != ior_OK)
  {
    errno = stdio_Err2ErrNo (IORes);
    return -1;
  }

  return 0;
}


// Old Turbo/Borland C/C++ doesn't have the S_I?USR defines:
#if !defined(S_IRUSR)
  #define S_IRUSR S_IREAD
  #define S_IWUSR S_IWRITE
  #define S_IXUSR S_IEXEC
#endif

int stat (
#if !(defined(__TURBOC__) && (__TURBOC__<=0x0296))
const
#endif
char *path, struct stat *statbuf)
{
  tIORes IORes;
  tFATFileStatus Status;
  struct tm BrokenDownTime;

  if (!stdio_IsFATFSInit)
    return -1;

  memset (statbuf, 0, sizeof(*statbuf));

  FAT_stat (&stdio_FATFS, &IORes, path, &Status);

  if (IORes != ior_OK)
  {
    errno = stdio_Err2ErrNo (IORes);
    return -1;
  }

  // RWX permissions:
  // directories will have S_IFDIR|S_IRUS|S_IXUSR mask,
  // regular files will have S_IFREG|S_IRUSR[|S_IWUSR] mask,
  // volume-IDs will have 0 mask:
  statbuf->st_mode =
    (Status.Attribute & dea_DIRECTORY) ? S_IFDIR|S_IRUSR|S_IXUSR : 
    (
      (Status.Attribute & dea_VOLUME_ID) ? 0 :
      (S_IFREG | S_IRUSR | ((Status.Attribute & dea_READ_ONLY) ? 0 : S_IWUSR))
    );
  // in DOS FAT only disk number can serve as device ID:
  statbuf->st_dev =
    statbuf->st_rdev = Status.DiskNo; // +1 ???

  statbuf->st_nlink = 1;

  statbuf->st_size = Status.FileSize;

  BrokenDownTime.tm_year = Status.Year - 1900;
  BrokenDownTime.tm_mon  = Status.Month - 1;
  BrokenDownTime.tm_mday = Status.Day;
  BrokenDownTime.tm_hour = Status.Hour;
  BrokenDownTime.tm_min  = Status.Minute;
  BrokenDownTime.tm_sec  = Status.Second;

  // DOS FAT isn't UNIX-like FS, so all the file times are the same as
  // the time of the last modification:
  statbuf->st_atime =
    statbuf->st_ctime =
      statbuf->st_mtime = mktime (&BrokenDownTime);

  return 0;
}

int chmod (const char *path, int amode)
{
  tIORes IORes;
  tModeT mode;

  if (!stdio_IsFATFSInit)
    return -1;

  if (amode & S_IWUSR)
    mode = mt_S_IWUSR;
  else
    mode = (tModeT)0;

  FAT_chmod (&stdio_FATFS, &IORes, path, mode);

  if (IORes != ior_OK)
  {
    errno = stdio_Err2ErrNo (IORes);
    return -1;
  }

  return 0;
}

#if defined(__TURBOC__) && (__TURBOC__ <= 0x0296)
// Turbo/Borland C/C++ earlier than 3.10 doesn't have utime():
struct utimbuf {
  time_t actime;
  time_t modtime;
};
#elif defined(__WATCOMC__)
// <Open> Watcom C/C++ has utime() in different header file:
  #include <sys/utime.h>
#else
  #include <utime.h>
#endif

int utime (
#if !defined(__TURBOC__)
const
#endif
char *path,
#if !defined(__TURBOC__)
const
#endif
struct utimbuf *times)
{
  tIORes IORes;
  struct tm *pBrokenDownTime;

  if (!stdio_IsFATFSInit)
    return -1;

  pBrokenDownTime = localtime (&times->modtime);
  FAT_utime
  (
    &stdio_FATFS, &IORes, path,
    pBrokenDownTime->tm_year + 1900,
    pBrokenDownTime->tm_mon + 1,
    pBrokenDownTime->tm_mday,
    pBrokenDownTime->tm_hour,
    pBrokenDownTime->tm_min,
    pBrokenDownTime->tm_sec
  );

  if (IORes != ior_OK)
  {
    errno = stdio_Err2ErrNo (IORes);
    return -1;
  }

  return 0;
}

tFATFile* _OpenFileSearch (tFATFS* pFATFS,
                           tReadDirState* pReadDirState,
                           tpOpenFilesMatchFxn pOpenFilesMatch)
{
  tFATFILE* pFile = stdio_pLastFile;

  if (!stdio_IsFATFSInit)
    return NULL;

  while (pFile != NULL)
  {
    if (pFile->IsSearchable)
    {
      if (pOpenFilesMatch (pFATFS, &pFile->FATFile, pReadDirState))
        return &pFile->FATFile;
    }
    pFile = pFile->pPrev;
  }

  return NULL;
}
