#include <string.h>
#include <stdio.h>
#include <errno.h>
#include <stdarg.h>
#include "mystdio.h"
#include "FAT.h"
#include "int13h.h"

#define ONLY_FLOPPIES_SUPPORTED 0
#define PRINT_RAW_FAT_ERORR_CODES 0

#ifdef __cplusplus
extern "C" {
#endif
extern void _GetDateAndTime (unsigned* pYear, unsigned* pMonth, unsigned* pDay,
                             unsigned* pHour, unsigned* pMinute, unsigned* pSecond);
extern tFATFile* _OpenFileSearch (tFATFS* pFATFS,
                                  tReadDirState* pReadDirState,
                                  tpOpenFilesMatchFxn pOpenFilesMatch);
#ifdef __cplusplus
}
#endif


tFATFS stdio_FATFS;
int stdio_IsFATFSInit = 0;

// Watcom C/C++ defines errno as a macro expanding to a value of an integer,
// whose address is returned by the function __get_errno_ptr():
#if defined(__WATCOMC__) && defined(errno)
int* __get_errno_ptr()
{
  static int __errno = 0;
  return &__errno;
}
#else
int errno = 0;
#endif

int stdio_Err2ErrNo (tIORes Res)
{
#if PRINT_RAW_FAT_ERORR_CODES
  if (Res != ior_OK)
    printf ("err=%u\n", (uint)Res);
#endif
  switch (Res)
  {
    case ior_OK:                // OK (no error)
      return 0;
    case ior_NOT_FOUND:         // file or directory not found
    case ior_BAD_NAME:          // attempted to find/create file or dir with bad name
      return ENOENT;
    case ior_DIR_FOUND:         // directory found
      return EISDIR;
    case ior_FILE_FOUND:        // file found
      return ENOTDIR;
    case ior_ALREADY_EXISTS:    // file or directory already exists by that name
      return EEXIST;
    case ior_PROHIBITED:        // permission denied, not enough rights, etc
      return EACCES;
    case ior_NO_SPACE:          // ran out of free disk space
      return ENOSPC;
    case ior_ERR_IO:            // input/output error (read/write/etc)
      return EIO;
    case ior_ERR_MEDIA_CHANGED: // media was changed, can't continue operation
      return ENODEV;
    case ior_NO_MEM:            // run out of RAM or null buffer pointer
      return ENOMEM;
    case ior_ARGUMENT_OUT_OF_RANGE:    // argument is out of range
      return ERANGE;
    case ior_INVALID_ARGUMENT:  // argument is invalid
      return EINVAL;
    default:
      return EFAULT;
  }
}

static tIORes ReadSectorLBA (tFATFS* pFATFS,
                             tFATBPB* pBPB, uint8 Drive, uint32 LBA,
                             uint8 SectorCount, void *pBuf)
{
  (void)pFATFS;
  return BIOS_ReadSectorLBA (pBPB, Drive, LBA, SectorCount, (void far*)pBuf);
}

static tIORes WriteSectorLBA (tFATFS* pFATFS,
                              tFATBPB* pBPB, uint8 Drive, uint32 LBA,
                              uint8 SectorCount, const void *pBuf)
{
  (void)pFATFS;
  return BIOS_WriteSectorLBA (pBPB, Drive, LBA, SectorCount, (const void far*)pBuf);
}

static int DoesDriveExist (tFATFS* pFATFS, uint8 Drive)
{
  (void)pFATFS;
#if ONLY_FLOPPIES_SUPPORTED
  if (Drive >= 0x80)
    return bdt_NO_DRIVE;
#endif
  return BIOS_GetDriveType (Drive) != bdt_NO_DRIVE;
}

static void GetDateAndTime (tFATFS* pFATFS,
                            unsigned* pYear, unsigned* pMonth, unsigned* pDay,
                            unsigned* pHour, unsigned* pMinute, unsigned* pSecond)
{
  (void)pFATFS;
  _GetDateAndTime (pYear, pMonth, pDay,
                   pHour, pMinute, pSecond);
}

static void _printf (tFATFS* pFATFS,
                     const char* format, ...)
{
#if 01
  (void)pFATFS;
  (void)format;
#else
  va_list args;

  (void)pFATFS;

  va_start (args, format);
  vprintf (format, args);

  va_end(args);
#endif
}

int stdio_InitFATFS()
{
  if (!stdio_IsFATFSInit)
    stdio_IsFATFSInit =
      FAT_Init
      (
        &stdio_FATFS,
        &DoesDriveExist,
        &ReadSectorLBA,
        &WriteSectorLBA,
        &GetDateAndTime,
        &_OpenFileSearch,
        &_printf
      );
  return stdio_IsFATFSInit;
}

int stdio_FlushFATFS()
{
  int res = 1;
  if (!stdio_IsFATFSInit)
    return 0;

  // First, flush all open files:
  flushall();

  if (!FAT_Flush (&stdio_FATFS))
    res = 0;

  return res;
}

int stdio_DoneFATFS()
{
  int res = 1;

  if (!stdio_IsFATFSInit)
    return 0;

  // First, close all open files:
  if (fcloseall() == EOF)
    res = 0;

  if (!FAT_Done (&stdio_FATFS))
    res = 0;

  return res;
}

#if 0
int main()
{
  FILE* f;
  char aString[256];
  stdio_InitFATFS();

  f = fopen ("c:\\autoexec.bat", "r");
  if (f == NULL)
    goto lend;

  while (fgets (aString, sizeof(aString), f) != NULL)
    printf ("%s", aString);

lend:
  if (f != NULL)
    fclose (f);

  stdio_DoneFATFS();
  return 0;
}
#endif
